//----------------------------------------------------------------------------------
//  lpf_v.c
//  LPF
//----------------------------------------------------------------------------------
#include <windows.h>
#include <mmintrin.h>
#include <string.h>

#include "filter.h"
#include "lpf_v.h"

#define CLIP(x, min, max) ((x)>(max) ? (max) : ((x)<(min) ? (min) : (x)))

// 3tap LPF
// ht Jn, hbAI荂
int lpf_v_3(PIXEL_YC *dst, PIXEL_YC *src, int h, int ht, int hb, int max_w, int w)
{
  int i,j;
  const int max_w3 = max_w * 3;
  short *src0, *src1, *src2, *dst0;
  short tmp;

  w *= (sizeof(PIXEL_YC) / sizeof(short));
  hb = h - hb;
  hb = CLIP(hb, 0, h);
  ht = CLIP(ht, 0, h);

  for(i=0; i<h; i++){
    dst0 = (short *)dst;
    src1 = (short *)src;
    
    if(i<ht || i>hb){
      memcpy(dst0, src1, w * sizeof(PIXEL_YC));
    }else{ 
      src0 = src1 + (i>ht     ? -max_w3 : 0);
      src2 = src1 + (i<(hb-1) ?  max_w3 : 0);
      
      for(j=0; j<w; ++j){
        tmp  = *(src0 + j) + *(src2 + j);
        tmp += *(src1 + j) * 2 + 2;
        tmp >>= 2;
        *dst0++ = (short)tmp;
      }
    }
    dst += max_w;
    src += max_w;
  }

  return 0;
}

// 5tap LPF
int lpf_v_5(PIXEL_YC *dst, PIXEL_YC *src, int h, int ht, int hb, int max_w, int w)
{
  int i,j;
  const int max_w3 = max_w * 3;
  short *src0, *src1, *src2, *src3, *src4, *dst0;
  int tmp;

  w *= (sizeof(PIXEL_YC) / sizeof(short));
  hb = h - hb;
  hb = CLIP(hb, 0, h);
  ht = CLIP(ht, 0, h);

  for(i=0; i<h; i++){
    dst0 = (short *)dst;
    src2 = (short *)src;

    if(i<ht || i>hb){
      memcpy(dst0, src2, max_w * sizeof(PIXEL_YC));
    }else{
      src1 = src2 + (i>ht     ? -max_w3 : 0);
      src0 = src1 + ((i-1)>ht ? -max_w3 : 0);
      src3 = src2 + (i<(hb-1) ?  max_w3 : 0);
      src4 = src3 + (i<(hb-2) ?  max_w3 : 0);
    
      for(j=0; j<w; ++j){
        tmp  = (*(src0+j) + *(src4+j)) * -2606;
        tmp += (*(src1+j) + *(src3+j)) *  8192;
        tmp += *(src2+j) * 21596 + (1<<14);
        *dst0++ = (short)(tmp >> 15);
      }
    }
    dst += max_w;
    src += max_w;
  }

  return 0;
}

// 7tap LPF
int lpf_v_7(PIXEL_YC *dst, PIXEL_YC *src, int h, int ht, int hb, int max_w, int w)
{
  int i,j;
  const int max_w3 = max_w * 3;
  short *src0, *src1, *src2, *src3, *src4, *src5, *src6, *dst0;
  int tmp;

  w *= (sizeof(PIXEL_YC) / sizeof(short));
  hb = h - hb;
  hb = CLIP(hb, 0, h);
  ht = CLIP(ht, 0, h);

  for(i=0; i<h; i++){
    dst0 = (short *)dst;
    src3 = (short *)src;
    dst += max_w;
    src += max_w;
    
    if(i<ht || i>hb){
      memcpy(dst0, src3, max_w * sizeof(PIXEL_YC));
      continue;
    }else if(i<ht+3 || i>hb-4){
      src2 = src3 + (i>ht     ? -max_w3 : 0);
      src1 = src2 + ((i-1)>ht ? -max_w3 : 0);
      src0 = src1 + ((i-2)>ht ? -max_w3 : 0);
      src4 = src3 + (i<(hb-1) ?  max_w3 : 0);
      src5 = src4 + (i<(hb-2) ?  max_w3 : 0);
      src6 = src5 + (i<(hb-3) ?  max_w3 : 0);
    }else{
      src2 = src3 - max_w3;
      src1 = src2 - max_w3;
      src0 = src1 - max_w3;
      src4 = src3 + max_w3;
      src5 = src4 + max_w3;
      src6 = src5 + max_w3;
    }
      
    for(j=0; j<w; ++j){
      tmp  = (*(src0+j) + *(src6+j)) *  1315;
      tmp += (*(src1+j) + *(src5+j)) * -4000;
      tmp += (*(src2+j) + *(src4+j)) *  6876;
      tmp += *(src3+j) * 24386 + (1<<14);
      *dst0++ = (short)(tmp >> 15);
    }
  }

  return 0;
}

// 9tap LPF
int lpf_v_9(PIXEL_YC *dst, PIXEL_YC *src, int h, int ht, int hb, int max_w, int w)
{
  int i,j;
  const int max_w3 = max_w * 3;
  short *src0, *src1, *src2, *src3, *src4, *src5, *src6, *src7, *src8, *dst0;
  int tmp;

  w *= (sizeof(PIXEL_YC) / sizeof(short));
  hb = h - hb;
  hb = CLIP(hb, 0, h);
  ht = CLIP(ht, 0, h);

  for(i=0; i<h; i++){
    dst0 = (short *)dst;
    src4 = (short *)src;
    dst += max_w;
    src += max_w;
    
    if(i<ht || i>hb){
      memcpy(dst0, src4, max_w * sizeof(PIXEL_YC));
      continue;
    }else if(i<ht+4 || i>hb-5){
      src3 = src4 + (i>ht     ? -max_w3 : 0);
      src2 = src3 + ((i-1)>ht ? -max_w3 : 0);
      src1 = src2 + ((i-2)>ht ? -max_w3 : 0);
      src0 = src1 + ((i-3)>ht ? -max_w3 : 0);
      src5 = src4 + (i<(hb-1) ?  max_w3 : 0);
      src6 = src5 + (i<(hb-2) ?  max_w3 : 0);
      src7 = src6 + (i<(hb-3) ?  max_w3 : 0);
      src8 = src7 + (i<(hb-4) ?  max_w3 : 0);
    }else{
      src3 = src4 - max_w3;
      src2 = src3 - max_w3;
      src1 = src2 - max_w3;
      src0 = src1 - max_w3;
      src5 = src4 + max_w3;
      src6 = src5 + max_w3;
      src7 = src6 + max_w3;
      src8 = src7 + max_w3;
    }
    
    for(j=0; j<w; ++j){
      tmp  = (*(src0+j) + *(src8+j)) *  -766;
      tmp += (*(src1+j) + *(src7+j)) *  2285;
      tmp += (*(src2+j) + *(src6+j)) * -4237;
      tmp += (*(src3+j) + *(src5+j)) *  5907;
      tmp += *(src4+j) * 26390 + (1<<14);
      *dst0++ = (short)(tmp >> 15);
    }
  }

  return 0;
}

// 11tap LPF
int lpf_v_11(PIXEL_YC *dst, PIXEL_YC *src, int h, int ht, int hb, int max_w, int w)
{
  int i,j;
  const int max_w3 = max_w * 3;
  short *src0, *src1, *src2, *src3, *src4, *src5;
  short *src6, *src7, *src8, *src9, *srca, *dst0;
  int tmp;

  w *= (sizeof(PIXEL_YC) / sizeof(short));
  hb = h - hb;
  hb = CLIP(hb, 0, h);
  ht = CLIP(ht, 0, h);

  for(i=0; i<h; i++){
    dst0 = (short *)dst;
    src5 = (short *)src;
    dst += max_w;
    src += max_w;
    
    if(i<ht || i>hb){
      memcpy(dst0, src5, max_w * sizeof(PIXEL_YC));
      continue;
    }else if(i<ht+5 || i>hb-6){
      src4 = src5 + (i>ht     ? -max_w3 : 0);
      src3 = src4 + ((i-1)>ht ? -max_w3 : 0);
      src2 = src3 + ((i-2)>ht ? -max_w3 : 0);
      src1 = src2 + ((i-3)>ht ? -max_w3 : 0);
      src0 = src1 + ((i-4)>ht ? -max_w3 : 0);
      src6 = src5 + (i<(hb-1) ?  max_w3 : 0);
      src7 = src6 + (i<(hb-2) ?  max_w3 : 0);
      src8 = src7 + (i<(hb-3) ?  max_w3 : 0);
      src9 = src8 + (i<(hb-4) ?  max_w3 : 0);
      srca = src9 + (i<(hb-5) ?  max_w3 : 0);
    }else{
      src4 = src5 - max_w3;
      src3 = src4 - max_w3;
      src2 = src3 - max_w3;
      src1 = src2 - max_w3;
      src0 = src1 - max_w3;
      src6 = src5 + max_w3;
      src7 = src6 + max_w3;
      src8 = src7 + max_w3;
      src9 = src8 + max_w3;
      srca = src9 + max_w3;
    }

    for(j=0; j<w; ++j){
      tmp  = (*(src0+j) + *(srca+j)) *   518;
      tmp += (*(src1+j) + *(src9+j)) * -1442;
      tmp += (*(src2+j) + *(src8+j)) *  2693;
      tmp += (*(src3+j) + *(src7+j)) * -3993;
      tmp += (*(src4+j) + *(src6+j)) *  4980;
      tmp += *(src5+j) * 27256 + (1<<14);
      *dst0++ = (short)(tmp >> 15);
    }
  }

  return 0;
}


// 13tap LPF
int lpf_v_13(PIXEL_YC *dst, PIXEL_YC *src, int h, int ht, int hb, int max_w, int w)
{
  int i,j;
  const int max_w3 = max_w * 3;
  short *src0, *src1, *src2, *src3, *src4, *src5, *src6;
  short *src7, *src8, *src9, *srca, *srcb, *srcc, *dst0;
  int tmp;

  w *= (sizeof(PIXEL_YC) / sizeof(short));
  hb = h - hb;
  hb = CLIP(hb, 0, h);
  ht = CLIP(ht, 0, h);

  for(i=0; i<h; i++){
    dst0 = (short *)dst;
    src6 = (short *)src;
    dst += max_w;
    src += max_w;
    
    if(i<ht || i>hb){
      memcpy(dst0, src6, max_w * sizeof(PIXEL_YC));
      continue;
    }else if(i<ht+6 || i>hb-7){
      src5 = src6 + (i>ht     ? -max_w3 : 0);
      src4 = src5 + ((i-1)>ht ? -max_w3 : 0);
      src3 = src4 + ((i-2)>ht ? -max_w3 : 0);
      src2 = src3 + ((i-3)>ht ? -max_w3 : 0);
      src1 = src2 + ((i-4)>ht ? -max_w3 : 0);
      src0 = src1 + ((i-5)>ht ? -max_w3 : 0);
      src7 = src6 + (i<(hb-1) ?  max_w3 : 0);
      src8 = src7 + (i<(hb-2) ?  max_w3 : 0);
      src9 = src8 + (i<(hb-3) ?  max_w3 : 0);
      srca = src9 + (i<(hb-4) ?  max_w3 : 0);
      srcb = srca + (i<(hb-5) ?  max_w3 : 0);
      srcc = srcb + (i<(hb-6) ?  max_w3 : 0);
    }else{
      src5 = src6 - max_w3;
      src4 = src5 - max_w3;
      src3 = src4 - max_w3;
      src2 = src3 - max_w3;
      src1 = src2 - max_w3;
      src0 = src1 - max_w3;
      src7 = src6 + max_w3;
      src8 = src7 + max_w3;
      src9 = src8 + max_w3;
      srca = src9 + max_w3;
      srcb = srca + max_w3;
      srcc = srcb + max_w3;
    }
    
    for(j=0; j<w; ++j){
      tmp  = (*(src0+j) + *(srcc+j)) *  -363;
      tmp += (*(src1+j) + *(srcb+j)) *   981;
      tmp += (*(src2+j) + *(srca+j)) * -1841;
      tmp += (*(src3+j) + *(src9+j)) *  2822;
      tmp += (*(src4+j) + *(src8+j)) * -3737;
      tmp += (*(src5+j) + *(src7+j)) *  4389;
      tmp += *(src6+j) * 28266 + (1<<14);
      *dst0++ = (short)(tmp >> 15);
    }
  }

  return 0;
}

// 15tap LPF
int lpf_v_15(PIXEL_YC *dst, PIXEL_YC *src, int h, int ht, int hb, int max_w, int w)
{
  int i,j;
  const int max_w3 = max_w * 3;
  short *src0, *src1, *src2, *src3, *src4, *src5, *src6, *src7;
  short *src8, *src9, *srca, *srcb, *srcc, *srcd, *srce, *dst0;
  int tmp;

  w *= (sizeof(PIXEL_YC) / sizeof(short));
  hb = h - hb;
  hb = CLIP(hb, 0, h);
  ht = CLIP(ht, 0, h);

  for(i=0; i<h; i++){
    dst0 = (short *)dst;
    src7 = (short *)src;
    dst += max_w;
    src += max_w;
    
    if(i<ht || i>hb){
      memcpy(dst0, src7, max_w * sizeof(PIXEL_YC));
      continue;
    }else if(i<ht+7 || i>hb-8){
      src6 = src7 + (i>ht     ? -max_w3 : 0);
      src5 = src6 + ((i-1)>ht ? -max_w3 : 0);
      src4 = src5 + ((i-2)>ht ? -max_w3 : 0);
      src3 = src4 + ((i-3)>ht ? -max_w3 : 0);
      src2 = src3 + ((i-4)>ht ? -max_w3 : 0);
      src1 = src2 + ((i-5)>ht ? -max_w3 : 0);
      src0 = src1 + ((i-6)>ht ? -max_w3 : 0);
      src8 = src7 + (i<(hb-1) ?  max_w3 : 0);
      src9 = src8 + (i<(hb-2) ?  max_w3 : 0);
      srca = src9 + (i<(hb-3) ?  max_w3 : 0);
      srcb = srca + (i<(hb-4) ?  max_w3 : 0);
      srcc = srcb + (i<(hb-5) ?  max_w3 : 0);
      srcd = srcc + (i<(hb-6) ?  max_w3 : 0);
      srce = srcd + (i<(hb-7) ?  max_w3 : 0);
    }else{
      src6 = src7 - max_w3;
      src5 = src6 - max_w3;
      src4 = src5 - max_w3;
      src3 = src4 - max_w3;
      src2 = src3 - max_w3;
      src1 = src2 - max_w3;
      src0 = src1 + max_w3;
      src8 = src7 + max_w3;
      src9 = src8 + max_w3;
      srca = src9 + max_w3;
      srcb = srca + max_w3;
      srcc = srcb + max_w3;
      srcd = srcc + max_w3;
      srce = srcd + max_w3;
    }
    
    for(j=0; j<w; ++j){
      tmp  = (*(src0+j) + *(srce+j)) *   265;
      tmp += (*(src1+j) + *(srcd+j)) *  -695;
      tmp += (*(src2+j) + *(srcc+j)) *  1299;
      tmp += (*(src3+j) + *(srcb+j)) * -2021;
      tmp += (*(src4+j) + *(srca+j)) *  2766;
      tmp += (*(src5+j) + *(src9+j)) * -3417;
      tmp += (*(src6+j) + *(src8+j)) *  3862;
      tmp += *(src7+j) * 28650 + (1<<14);
      *dst0++ = (short)(tmp >> 15);
    }
  }

  return 0;
}

// 3tap LPF MMX
// ht Jn, hbAI荂
int lpf_v_3_mmx(PIXEL_YC *dst, PIXEL_YC *src, int h, int ht, int hb, int max_w, int w)
{
  int i,j;
  const int max_w3 = max_w * 3;
  short *src0, *src1, *src2, *dst0;
  short tmp;
  __m64 s0, s1, c0;

  c0 = _mm_set1_pi16(2);
  w *= (sizeof(PIXEL_YC) / sizeof(short));
  hb = h - hb;
  hb = CLIP(hb, 0, h);
  ht = CLIP(ht, 0, h);

  for(i=0; i<h; i++){
    dst0 = (short *)dst;
    src1 = (short *)src;
    
    if(i<ht || i>hb){
      memcpy(dst0, src1, w * sizeof(PIXEL_YC));
    }else{ 
      src0 = src1 + (i>ht     ? -max_w3 : 0);
      src2 = src1 + (i<(hb-1) ?  max_w3 : 0);
      
      for(j=w; j>3; j-=4){
        s0 = _mm_adds_pi16(*(__m64 *)src0, *(__m64 *)src2);
        s1 = _mm_slli_pi16(*(__m64 *)src1, 1);
        s0 = _mm_adds_pi16(s0, c0);
        s1 = _mm_adds_pi16(s1, s0);
        *(__m64 *)dst0 = _mm_srai_pi16(s1, 2);

        src0 += 4;
        src1 += 4;
        src2 += 4;
        dst0 += 4;
      }
      //c
      for( ; j>0; --j){
        tmp  = *src0 + *src2;
        tmp += *src1 * 2 + 2;
        tmp >>= 2;
        *dst0++ = (short)tmp;

        src0++;
        src1++;
        src2++;
      }
    }
    dst += max_w;
    src += max_w;
  }

  _mm_empty();
  
  return 0;
}

// 5tap LPF MMX
int lpf_v_5_mmx(PIXEL_YC *dst, PIXEL_YC *src, int h, int ht, int hb, int max_w, int w)
{
  int i,j;
  const int max_w3 = max_w * 3;
  short *src0, *src1, *src2, *src3, *src4, *dst0;
  int tmp;
  __m64 s0, s1, s2, s3, c0, c1, c2;
  
  c0 = _mm_set_pi16(-2606, 8192, -2606, 8192);
  c1 = _mm_set_pi16(0, 21596, 0, 21596);
  c2 = _mm_set1_pi32(1<<14);
  hb = h - hb;
  hb = CLIP(hb, 0, h);
  ht = CLIP(ht, 0, h);
  w  = w * (sizeof(PIXEL_YC) / sizeof(short));

  for(i=0; i<h; i++){
    dst0 = (short *)dst;
    src2 = (short *)src;

    if(i<ht || i>hb){
      memcpy(dst0, src2, max_w * sizeof(PIXEL_YC));
    }else{
      src1 = src2 + (i>ht     ? -max_w3 : 0);
      src0 = src1 + ((i-1)>ht ? -max_w3 : 0);
      src3 = src2 + (i<(hb-1) ?  max_w3 : 0);
      src4 = src3 + (i<(hb-2) ?  max_w3 : 0);
      
      for(j=w; j>3; j-=4){
        s2 = _mm_adds_pi16(*(__m64 *)src0, *(__m64 *)src4);
        s1 = _mm_adds_pi16(*(__m64 *)src1, *(__m64 *)src3);
        s0 = s1;
        
        s1 = _mm_unpackhi_pi16(s1, s2);
        s0 = _mm_unpacklo_pi16(s0, s2);
        
        s1 = _mm_madd_pi16(s1, c0);
        s0 = _mm_madd_pi16(s0, c0);

        s0 = _mm_add_pi32(s0, c2);
        s1 = _mm_add_pi32(s1, c2);

        s2 = *(__m64 *)src2;
        s3 = s2;
        s3 = _mm_unpackhi_pi16(s3, c0);
        s2 = _mm_unpacklo_pi16(s2, c0);
        
        s2 = _mm_madd_pi16(s2, c1);
        s3 = _mm_madd_pi16(s3, c1);
        
        s0 = _mm_add_pi32(s0, s2);
        s1 = _mm_add_pi32(s1, s3);
        
        s0 = _mm_srai_pi32(s0, 15);
        s1 = _mm_srai_pi32(s1, 15);
        
        *(__m64 *)dst0 = _mm_packs_pi32(s0, s1);
        
        src0 += 4;
        src1 += 4;
        src2 += 4;
        src3 += 4;
        src4 += 4;
        dst0 += 4;
      }
      //̂
      for( ; j>0; --j){
        tmp  = (*src0 + *src4) * -2606;
        tmp += (*src1 + *src3) *  8192;
        tmp += *src2 * 21596 + (1<<14);
        *dst0++ = (short)(tmp >> 15);

        src0++;
        src1++;
        src2++;
        src3++;
        src4++;
      }
    }
    dst += max_w;
    src += max_w;
  }
  
  _mm_empty();
  
  return 0;
}


// 7tap LPF MMX
int lpf_v_7_mmx(PIXEL_YC *dst, PIXEL_YC *src, int h, int ht, int hb, int max_w, int w)
{
  int i,j;
  const int max_w3 = max_w * 3;
  short *src0, *src1, *src2, *src3, *src4, *src5, *src6, *dst0;
  int tmp;
  __m64 s0, s1, s2, s3, s4, c0, c1, c2;
  
  c0 = _mm_set_pi16(1315, -4000, 1315, -4000);
  c1 = _mm_set_pi16(24386, 6876, 24386, 6876);
  c2 = _mm_set1_pi32(1<<14);
  hb = h - hb;
  hb = CLIP(hb, 0, h);
  ht = CLIP(ht, 0, h);
  w  = w * (sizeof(PIXEL_YC) / sizeof(short));

  for(i=0; i<h; i++){
    dst0 = (short *)dst;
    src3 = (short *)src;
    dst += max_w;
    src += max_w;
    
    if(i<ht || i>hb){
      memcpy(dst0, src3, max_w * sizeof(PIXEL_YC));
      continue;
    }else if(i<ht+3 || i>hb-4){
      src2 = src3 + (i>ht     ? -max_w3 : 0);
      src1 = src2 + ((i-1)>ht ? -max_w3 : 0);
      src0 = src1 + ((i-2)>ht ? -max_w3 : 0);
      src4 = src3 + (i<(hb-1) ?  max_w3 : 0);
      src5 = src4 + (i<(hb-2) ?  max_w3 : 0);
      src6 = src5 + (i<(hb-3) ?  max_w3 : 0);
    }else{
      src2 = src3 - max_w3;
      src1 = src2 - max_w3;
      src0 = src1 - max_w3;
      src4 = src3 + max_w3;
      src5 = src4 + max_w3;
      src6 = src5 + max_w3;
    }
    
    for(j=w; j>3; j-=4){
      s2 = _mm_adds_pi16(*(__m64 *)src0, *(__m64 *)src6); 
      s1 = _mm_adds_pi16(*(__m64 *)src1, *(__m64 *)src5);
      s0 = s1;
      
      s1 = _mm_unpackhi_pi16(s1, s2);
      s0 = _mm_unpacklo_pi16(s0, s2);

      s0 = _mm_madd_pi16(s0, c0);
      s1 = _mm_madd_pi16(s1, c0);

      s0 = _mm_add_pi32(s0, c2);
      s1 = _mm_add_pi32(s1, c2);
      

      s4 = _mm_adds_pi16(*(__m64 *)src2, *(__m64 *)src4);
      s3 = *(__m64 *)src3;
      s2 = s4;
      s4 = _mm_unpackhi_pi16(s4, s3);
      s2 = _mm_unpacklo_pi16(s2, s3);
      
      s4 = _mm_madd_pi16(s4, c1);
      s2 = _mm_madd_pi16(s2, c1);

      s1 = _mm_add_pi32(s1, s4);
      s0 = _mm_add_pi32(s0, s2);
      
      s1 = _mm_srai_pi32(s1, 15);
      s0 = _mm_srai_pi32(s0, 15);

      *(__m64 *)dst0 = _mm_packs_pi32(s0, s1);

      src0 += 4;
      src1 += 4;
      src2 += 4;
      src3 += 4;
      src4 += 4;
      src5 += 4;
      src6 += 4;
      dst0 += 4;
    }
    //c
    for( ; j>0; --j){
      tmp  = (*src0 + *src6) *  1315;
      tmp += (*src1 + *src5) * -4000;
      tmp += (*src2 + *src4) *  6876;
      tmp += *src3 * 24386 + (1<<14);
      *dst0++ = (short)(tmp >> 15);

      src0++;
      src1++;
      src2++;
      src3++;
      src4++;
      src5++;
      src6++;
    }
  }

  _mm_empty();

  return 0;
}

// 9tap LPF MMX
int lpf_v_9_mmx(PIXEL_YC *dst, PIXEL_YC *src, int h, int ht, int hb, int max_w, int w)
{
  int i,j;
  const int max_w3 = max_w * 3;
  short *src0, *src1, *src2, *src3, *src4, *src5, *src6, *src7, *src8, *dst0;
  int tmp;
  __m64 s0, s1, s2, s3, s4, c0, c1, c2, c3;
  
  c0 = _mm_set_pi16(-766, 2285, -766, 2285);
  c1 = _mm_set_pi16(-4237, 5907, -4237, 5907);
  c2 = _mm_set_pi16(26390, 1<<14, 26390, 1<<14);
  c3 = _mm_set1_pi16(1);
  hb = h - hb;
  hb = CLIP(hb, 0, h);
  ht = CLIP(ht, 0, h);
  w  = w * (sizeof(PIXEL_YC) / sizeof(short));
  
  for(i=0; i<h; i++){
    dst0 = (short *)dst;
    src4 = (short *)src;
    dst += max_w;
    src += max_w;
    
    if(i<ht || i>hb){
      memcpy(dst0, src4, max_w * sizeof(PIXEL_YC));
      continue;
    }else if(i<ht+4 || i>hb-5){
      src3 = src4 + (i>ht     ? -max_w3 : 0);
      src2 = src3 + ((i-1)>ht ? -max_w3 : 0);
      src1 = src2 + ((i-2)>ht ? -max_w3 : 0);
      src0 = src1 + ((i-3)>ht ? -max_w3 : 0);
      src5 = src4 + (i<(hb-1) ?  max_w3 : 0);
      src6 = src5 + (i<(hb-2) ?  max_w3 : 0);
      src7 = src6 + (i<(hb-3) ?  max_w3 : 0);
      src8 = src7 + (i<(hb-4) ?  max_w3 : 0);
    }else{
      src3 = src4 - max_w3;
      src2 = src3 - max_w3;
      src1 = src2 - max_w3;
      src0 = src1 - max_w3;
      src5 = src4 + max_w3;
      src6 = src5 + max_w3;
      src7 = src6 + max_w3;
      src8 = src7 + max_w3;
    }

    for(j=w; j>3; j-=4){
      s2 = _mm_adds_pi16(*(__m64 *)src0, *(__m64 *)src8);
      s1 = _mm_adds_pi16(*(__m64 *)src1, *(__m64 *)src7);
      s0 = s1;
      
      s0 = _mm_unpackhi_pi16(s0, s2);
      s1 = _mm_unpacklo_pi16(s1, s2);
      s0 = _mm_madd_pi16(s0, c0);
      s1 = _mm_madd_pi16(s1, c0);

      //
      s4 = _mm_adds_pi16(*(__m64 *)src2, *(__m64 *)src6);
      s3 = _mm_adds_pi16(*(__m64 *)src3, *(__m64 *)src5);
      s2 = s3;

      s2 = _mm_unpackhi_pi16(s2, s4);
      s3 = _mm_unpacklo_pi16(s3, s4);
      s2 = _mm_madd_pi16(s2, c1);
      s3 = _mm_madd_pi16(s3, c1);

      //
      s0 = _mm_add_pi32(s0, s2);
      s1 = _mm_add_pi32(s1, s3);

      //
      s4 = *(__m64 *)src4;
      s2 = s3 = c3;
      
      s2 = _mm_unpackhi_pi16(s2, s4);
      s3 = _mm_unpacklo_pi16(s3, s4);
      s2 = _mm_madd_pi16(s2, c2);
      s3 = _mm_madd_pi16(s3, c2);

      s0 = _mm_add_pi32(s0, s2);
      s1 = _mm_add_pi32(s1, s3);

      s0 = _mm_srai_pi32(s0, 15);
      s1 = _mm_srai_pi32(s1, 15);

      *(__m64 *)dst0 = _mm_packs_pi32(s1, s0);

      src0 += 4;
      src1 += 4;
      src2 += 4;
      src3 += 4;
      src4 += 4;
      src5 += 4;
      src6 += 4;
      src7 += 4;
      src8 += 4;     
      dst0 += 4;
    }
    //̂
    for( ; j>0; --j){
      tmp  = (*src0 + *src8) *  -766;
      tmp += (*src1 + *src7) *  2285;
      tmp += (*src2 + *src6) * -4237;
      tmp += (*src3 + *src5) *  5907;
      tmp += *src4 * 26390 + (1<<14);
      *dst0++ = (short)(tmp >> 15);
      
      src0++;
      src1++;
      src2++;
      src3++;
      src4++;
      src5++;
      src6++;
      src7++;
      src8++;      
    }
  }
  
  _mm_empty();
  
  return 0;
}

// 11tap LPF MMX
int lpf_v_11_mmx(PIXEL_YC *dst, PIXEL_YC *src, int h, int ht, int hb, int max_w, int w)
{
  int i,j;
  const int max_w3 = max_w * 3;
  short *src0, *src1, *src2, *src3, *src4, *src5;
  short *src6, *src7, *src8, *src9, *srca, *dst0;
  int tmp;
  __m64 s0, s1, s2, s3, s4, c0, c1, c2, c3;
  
  c0 = _mm_set_pi16(518, -1442, 518, -1442);
  c1 = _mm_set_pi16(2693, -3993, 2693, -3993);
  c2 = _mm_set_pi16(27256, 4980, 27256, 4980);
  c3 = _mm_set1_pi32(1<<14);
  hb = h - hb;
  hb = CLIP(hb, 0, h);
  ht = CLIP(ht, 0, h);
  w  = w * (sizeof(PIXEL_YC) / sizeof(short));

  for(i=0; i<h; i++){
    dst0 = (short *)dst;
    src5 = (short *)src;
    dst += max_w;
    src += max_w;
    
    if(i<ht || i>hb){
      memcpy(dst0, src5, max_w * sizeof(PIXEL_YC));
      continue;
    }else if(i<ht+5 || i>hb-6){
      src4 = src5 + (i>ht     ? -max_w3 : 0);
      src3 = src4 + ((i-1)>ht ? -max_w3 : 0);
      src2 = src3 + ((i-2)>ht ? -max_w3 : 0);
      src1 = src2 + ((i-3)>ht ? -max_w3 : 0);
      src0 = src1 + ((i-4)>ht ? -max_w3 : 0);
      src6 = src5 + (i<(hb-1) ?  max_w3 : 0);
      src7 = src6 + (i<(hb-2) ?  max_w3 : 0);
      src8 = src7 + (i<(hb-3) ?  max_w3 : 0);
      src9 = src8 + (i<(hb-4) ?  max_w3 : 0);
      srca = src9 + (i<(hb-5) ?  max_w3 : 0);
    }else{
      src4 = src5 - max_w3;
      src3 = src4 - max_w3;
      src2 = src3 - max_w3;
      src1 = src2 - max_w3;
      src0 = src1 - max_w3;
      src6 = src5 + max_w3;
      src7 = src6 + max_w3;
      src8 = src7 + max_w3;
      src9 = src8 + max_w3;
      srca = src9 + max_w3;
    }
    
    for(j=w; j>3; j-=4){
      //
      s2 = _mm_adds_pi16(*(__m64 *)src0, *(__m64 *)srca);
      s1 = _mm_adds_pi16(*(__m64 *)src1, *(__m64 *)src9);
      s0 = s1;

      s0 = _mm_unpackhi_pi16(s0, s2);
      s1 = _mm_unpacklo_pi16(s1, s2);
      s0 = _mm_madd_pi16(s0, c0);
      s1 = _mm_madd_pi16(s1, c0);

      s0 = _mm_add_pi32(s0, c3);
      s1 = _mm_add_pi32(s1, c3);
      
      //
      s4 = _mm_adds_pi16(*(__m64 *)src2, *(__m64 *)src8);
      s3 = _mm_adds_pi16(*(__m64 *)src3, *(__m64 *)src7);
      s2 = s3;
      
      s2 = _mm_unpackhi_pi16(s2, s4);
      s3 = _mm_unpacklo_pi16(s3, s4);
      s2 = _mm_madd_pi16(s2, c1);
      s3 = _mm_madd_pi16(s3, c1);

      s0 = _mm_add_pi32(s0, s2);
      s1 = _mm_add_pi32(s1, s3);

      //
      s3 = _mm_adds_pi16(*(__m64 *)src4, *(__m64 *)src6);
      s4 = *(__m64 *)src5;
      s2 = s3;
      s2 = _mm_unpackhi_pi16(s2, s4);
      s3 = _mm_unpacklo_pi16(s3, s4);
      
      s2 = _mm_madd_pi16(s2, c2);
      s3 = _mm_madd_pi16(s3, c2);

      s0 = _mm_add_pi32(s0, s2);
      s1 = _mm_add_pi32(s1, s3);

      s0 = _mm_srai_pi32(s0, 15);
      s1 = _mm_srai_pi32(s1, 15);

      *(__m64 *)dst0 = _mm_packs_pi32(s1, s0);

      src0 += 4;
      src1 += 4;
      src2 += 4;
      src3 += 4;
      src4 += 4;
      src5 += 4;
      src6 += 4;
      src7 += 4;
      src8 += 4;
      src9 += 4;
      srca += 4;
      dst0 += 4;
    }
    //̂
    for( ; j>0; --j){
      tmp  = (*src0 + *srca) *   518;
      tmp += (*src1 + *src9) * -1442;
      tmp += (*src2 + *src8) *  2693;
      tmp += (*src3 + *src7) * -3993;
      tmp += (*src4 + *src6) *  4980;
      tmp += *src5 * 27256 + (1<<14);
      *dst0++ = (short)(tmp >> 15);

      src0++;
      src1++;
      src2++;
      src3++;
      src4++;
      src5++;
      src6++;
      src7++;
      src8++;
      src9++;
      srca++;     
    }
  }

  _mm_empty();

  return 0;
}

// 13tap LPF MMX
int lpf_v_13_mmx(PIXEL_YC *dst, PIXEL_YC *src, int h, int ht, int hb, int max_w, int w)
{
  int i,j;
  const int max_w3 = max_w * 3;
  short *src0, *src1, *src2, *src3, *src4, *src5, *src6;
  short *src7, *src8, *src9, *srca, *srcb, *srcc, *dst0;
  int tmp;
  __m64 s0, s1, s2, s3, s4, c0, c1, c2, c3, c4;

  c0 = _mm_set_pi16(-363, 981, -363, 981);
  c1 = _mm_set_pi16(-1841, 2822, -1841, 2822);
  c2 = _mm_set_pi16(-3737, 4389, -3737, 4389);
  c3 = _mm_set_pi16(28266, 1<<14, 28266, 1<<14);
  c4 = _mm_set1_pi16(1);
  hb = h - hb;
  hb = CLIP(hb, 0, h);
  ht = CLIP(ht, 0, h);
  w  = w * (sizeof(PIXEL_YC) / sizeof(short));
  
  for(i=0; i<h; i++){
    dst0 = (short *)dst;
    src6 = (short *)src;
    dst += max_w;
    src += max_w;
    
    if(i<ht || i>hb){
      memcpy(dst0, src6, max_w * sizeof(PIXEL_YC));
      continue;
    }else if(i<ht+6 || i>hb-7){
      src5 = src6 + (i>ht     ? -max_w3 : 0);
      src4 = src5 + ((i-1)>ht ? -max_w3 : 0);
      src3 = src4 + ((i-2)>ht ? -max_w3 : 0);
      src2 = src3 + ((i-3)>ht ? -max_w3 : 0);
      src1 = src2 + ((i-4)>ht ? -max_w3 : 0);
      src0 = src1 + ((i-5)>ht ? -max_w3 : 0);
      src7 = src6 + (i<(hb-1) ?  max_w3 : 0);
      src8 = src7 + (i<(hb-2) ?  max_w3 : 0);
      src9 = src8 + (i<(hb-3) ?  max_w3 : 0);
      srca = src9 + (i<(hb-4) ?  max_w3 : 0);
      srcb = srca + (i<(hb-5) ?  max_w3 : 0);
      srcc = srcb + (i<(hb-6) ?  max_w3 : 0);
    }else{
      src5 = src6 - max_w3;
      src4 = src5 - max_w3;
      src3 = src4 - max_w3;
      src2 = src3 - max_w3;
      src1 = src2 - max_w3;
      src0 = src1 - max_w3;
      src7 = src6 + max_w3;
      src8 = src7 + max_w3;
      src9 = src8 + max_w3;
      srca = src9 + max_w3;
      srcb = srca + max_w3;
      srcc = srcb + max_w3;
    }
    
    for(j=w; j>3; j-=4){
      s2 = _mm_adds_pi16(*(__m64 *)src0, *(__m64 *)srcc);
      s1 = _mm_adds_pi16(*(__m64 *)src1, *(__m64 *)srcb);
      s0 = s1;

      s0 = _mm_unpackhi_pi16(s0, s2);
      s1 = _mm_unpacklo_pi16(s1, s2);

      s0 = _mm_madd_pi16(s0, c0);
      s1 = _mm_madd_pi16(s1, c0);

      //
      s4 = _mm_adds_pi16(*(__m64 *)src2, *(__m64 *)srca);
      s3 = _mm_adds_pi16(*(__m64 *)src3, *(__m64 *)src9);
      s2 = s3;

      s2 = _mm_unpackhi_pi16(s2, s4);
      s3 = _mm_unpacklo_pi16(s3, s4);

      s2 = _mm_madd_pi16(s2, c1);
      s3 = _mm_madd_pi16(s3, c1);
      
      s0 = _mm_add_pi32(s0, s2);
      s1 = _mm_add_pi32(s1, s3);

      //
      s4 = _mm_adds_pi16(*(__m64 *)src4, *(__m64 *)src8);
      s3 = _mm_adds_pi16(*(__m64 *)src5, *(__m64 *)src7);
      s2 = s3;

      s2 = _mm_unpackhi_pi16(s2, s4);
      s3 = _mm_unpacklo_pi16(s3, s4);

      s2 = _mm_madd_pi16(s2, c2);
      s3 = _mm_madd_pi16(s3, c2);
      
      s0 = _mm_add_pi32(s0, s2);
      s1 = _mm_add_pi32(s1, s3);

      //
      s4 = *(__m64 *)src6;
      s3 = s2 = c4;
      
      s2 = _mm_unpackhi_pi16(s2, s4);
      s3 = _mm_unpacklo_pi16(s3, s4);

      s2 = _mm_madd_pi16(s2, c3);
      s3 = _mm_madd_pi16(s3, c3);

      s0 = _mm_add_pi32(s0, s2);
      s1 = _mm_add_pi32(s1, s3);

      s0 = _mm_srai_pi32(s0, 15);
      s1 = _mm_srai_pi32(s1, 15);

      *(__m64 *)dst0 = _mm_packs_pi32(s1, s0);

      src0 += 4;
      src1 += 4;
      src2 += 4;
      src3 += 4;
      src4 += 4;
      src5 += 4;
      src6 += 4;
      src7 += 4;
      src8 += 4;
      src9 += 4;
      srca += 4;
      srcb += 4;
      srcc += 4;     
      dst0 += 4;
    }
    for( ; j>0; --j){
      tmp  = (*src0 + *srcc) *  -363;
      tmp += (*src1 + *srcb) *   981;
      tmp += (*src2 + *srca) * -1841;
      tmp += (*src3 + *src9) *  2822;
      tmp += (*src4 + *src8) * -3737;
      tmp += (*src5 + *src7) *  4389;
      tmp += *src6 * 28266 + (1<<14);
      *dst0++ = (short)(tmp >> 15);

      src0++;
      src1++;
      src2++;
      src3++;
      src4++;
      src5++;
      src6++;
      src7++;
      src8++;
      src9++;
      srca++;
      srcb++;
      srcc++;     
    }
  }

  _mm_empty();

  return 0;
}

// 15tap LPF MMX
int lpf_v_15_mmx(PIXEL_YC *dst, PIXEL_YC *src, int h, int ht, int hb, int max_w, int w)
{
  int i,j;
  const int max_w3 = max_w * 3;
  short *src0, *src1, *src2, *src3, *src4, *src5, *src6, *src7;
  short *src8, *src9, *srca, *srcb, *srcc, *srcd, *srce, *dst0;
  int tmp;
  __m64 s0, s1, s2, s3, s4, c0, c1, c2, c3, c4;

  c0 = _mm_set_pi16(265, -695, 265, -695);
  c1 = _mm_set_pi16(1299, -2021, 1299, -2021);
  c2 = _mm_set_pi16(2766, -3417, 2766, -3417);
  c3 = _mm_set_pi16(3862, 28650, 3862, 28650);
  c4 = _mm_set1_pi32(1<<14);
  hb = h - hb;
  hb = CLIP(hb, 0, h);
  ht = CLIP(ht, 0, h);
  w  = w * (sizeof(PIXEL_YC) / sizeof(short));
  
  for(i=0; i<h; i++){
    dst0 = (short *)dst;
    src7 = (short *)src;
    dst += max_w;
    src += max_w;
    
    if(i<ht || i>hb){
      memcpy(dst0, src7, max_w * sizeof(PIXEL_YC));
      continue;
    }else if(i<ht+7 || i>hb-8){
      src6 = src7 + (i>ht     ? -max_w3 : 0);
      src5 = src6 + ((i-1)>ht ? -max_w3 : 0);
      src4 = src5 + ((i-2)>ht ? -max_w3 : 0);
      src3 = src4 + ((i-3)>ht ? -max_w3 : 0);
      src2 = src3 + ((i-4)>ht ? -max_w3 : 0);
      src1 = src2 + ((i-5)>ht ? -max_w3 : 0);
      src0 = src1 + ((i-6)>ht ? -max_w3 : 0);
      src8 = src7 + (i<(hb-1) ?  max_w3 : 0);
      src9 = src8 + (i<(hb-2) ?  max_w3 : 0);
      srca = src9 + (i<(hb-3) ?  max_w3 : 0);
      srcb = srca + (i<(hb-4) ?  max_w3 : 0);
      srcc = srcb + (i<(hb-5) ?  max_w3 : 0);
      srcd = srcc + (i<(hb-6) ?  max_w3 : 0);
      srce = srcd + (i<(hb-7) ?  max_w3 : 0);
    }else{
      src6 = src7 - max_w3;
      src5 = src6 - max_w3;
      src4 = src5 - max_w3;
      src3 = src4 - max_w3;
      src2 = src3 - max_w3;
      src1 = src2 - max_w3;
      src0 = src1 + max_w3;
      src8 = src7 + max_w3;
      src9 = src8 + max_w3;
      srca = src9 + max_w3;
      srcb = srca + max_w3;
      srcc = srcb + max_w3;
      srcd = srcc + max_w3;
      srce = srcd + max_w3;
    }
    
    for(j=w; j>3; j-=4){
      s2 = _mm_adds_pi16(*(__m64 *)src0, *(__m64 *)srce);
      s1 = _mm_adds_pi16(*(__m64 *)src1, *(__m64 *)srcd);
      s0 = s1;

      s0 = _mm_unpackhi_pi16(s0, s2);
      s1 = _mm_unpacklo_pi16(s1, s2);

      s0 = _mm_madd_pi16(s0, c0);
      s1 = _mm_madd_pi16(s1, c0);

      s0 = _mm_add_pi32(s0, c4);
      s1 = _mm_add_pi32(s1, c4);
      
      //
      s4 = _mm_adds_pi16(*(__m64 *)src2, *(__m64 *)srcc);
      s3 = _mm_adds_pi16(*(__m64 *)src3, *(__m64 *)srcb);
      s2 = s3;
      
      s2 = _mm_unpackhi_pi16(s2, s4);
      s3 = _mm_unpacklo_pi16(s3, s4);

      s2 = _mm_madd_pi16(s2, c1);
      s3 = _mm_madd_pi16(s3, c1);

      s0 = _mm_add_pi32(s0, s2);
      s1 = _mm_add_pi32(s1, s3);

      //
      s4 = _mm_adds_pi16(*(__m64 *)src4, *(__m64 *)srca);
      s3 = _mm_adds_pi16(*(__m64 *)src5, *(__m64 *)src9);
      s2 = s3;
      
      s2 = _mm_unpackhi_pi16(s2, s4);
      s3 = _mm_unpacklo_pi16(s3, s4);

      s2 = _mm_madd_pi16(s2, c2);
      s3 = _mm_madd_pi16(s3, c2);

      s0 = _mm_add_pi32(s0, s2);
      s1 = _mm_add_pi32(s1, s3);

      //
      s4 = _mm_adds_pi16(*(__m64 *)src6, *(__m64 *)src8);
      s3 = *(__m64 *)src7;
      s2 = s3;
      
      s2 = _mm_unpackhi_pi16(s2, s4);
      s3 = _mm_unpacklo_pi16(s3, s4);
      
      s2 = _mm_madd_pi16(s2, c3);
      s3 = _mm_madd_pi16(s3, c3);

      s0 = _mm_add_pi32(s0, s2);
      s1 = _mm_add_pi32(s1, s3);

      s0 = _mm_srai_pi32(s0, 15);
      s1 = _mm_srai_pi32(s1, 15);

      *(__m64 *)dst0 = _mm_packs_pi32(s1, s0);

      src0 += 4;
      src1 += 4;
      src2 += 4;
      src3 += 4;
      src4 += 4;
      src5 += 4;
      src6 += 4;
      src7 += 4;
      src8 += 4;
      src9 += 4;
      srca += 4;
      srcb += 4;
      srcc += 4;
      srcd += 4;
      srce += 4;
      dst0 += 4;
    }
    for( ; j>0; --j){
      tmp  = (*src0 + *srce) *   265;
      tmp += (*src1 + *srcd) *  -695;
      tmp += (*src2 + *srcc) *  1299;
      tmp += (*src3 + *srcb) * -2021;
      tmp += (*src4 + *srca) *  2766;
      tmp += (*src5 + *src9) * -3417;
      tmp += (*src6 + *src8) *  3862;
      tmp += *src7 * 28650 + (1<<14);
      *dst0++ = (short)(tmp >> 15);

      src0++;
      src1++;
      src2++;
      src3++;
      src4++;
      src5++;
      src6++;
      src7++;
      src8++;
      src9++;
      srca++;
      srcb++;
      srcc++;
      srcd++;
      srce++;
    }
  }

  _mm_empty();

  return 0;
}

